#!/usr/bin/env python
'''
This tool runs remote commands on or more hosts using the SSH
protocol. It uses the paramiko package so you do not have to have ssh
installed.

Each host requires four items of information: hostname, port, username
and password but it is often the case that the same username,
password and port is used for all hosts. For secure environments every
username, password and port may be different. This tool handles both
cases.

The easiest approach is to provide a single username, password and
port for all hosts. This is typically the case for installations where
there is a role account that is used for certain maintenance
tasks. The example below shows how it works.

    EXAMPLE 01: same username, password and port for all hosts.

    $ # Run a simple command on several hosts using the same
    $ # username and password. The port is also the same but
    $ # this example does not show it being using (-P) because
    $ # assumes the default (port 22).
    $ rcmd.py -u admin -p "password" -H host1,host2 -c "pwd"

    $ # Here is another way to do the same thing.
    $ rcmd.py -u admin -p "password" -H host1 -H host2 -c "pwd"

The example above has the drawback of exposing the password on the
command line. That limitation can be addressed in two ways. You can
cat a protected file that contains the password or you can not specify
the -p option in which case you will be prompted for the password. The
example below shows how these options work.

    EXAMPLE 02: make the password entry a bit more secure

    $ # Run a simple command on several hosts using the same
    $ # username and password.
    $ rcmd.py -u admin -p "`cat passfile`" -H host1,host2 -c "pwd"

    $ # If you don't specify -p and you do specified -u, you will be
    % # prompted.
    $ rcmd.py -u admin -H host1,host2 -c "pwd"
    admin password:

If you are using a different username and password for each host there
are a couple of options. You can provide the username and password for
each host on the command line using the -D (--define) option or you
can define and use a login conf file in YAML format.

I generally prefer to use the login conf file because I can define a
single, protected conf file for all hosts and encrypt it. I then use
regular expressions to pick the hosts that are interesting.

The example below shows how to specify the username and password for
each host using the extended host format.

    EXAMPLE 03: username and passwords for each host on the CLI

    $ # Specify the username, password and port for each host.
    $ # If you specify 0 for the port, it will use the global
    $ # default value (-P) or 22 if no global was specified.
    $ # If you specify '' for either the username or password,
    $ # the associated global value will be used (-u or -p).
    $ rcmd.py -D 'host1' 'admin' 'password' 22 \\
              -D 'host2' 'admin' 'password' 22 \\
              -c pwd

As you can see, it can get verbose pretty quickly. Another way to do
the same thing is to define a login conf file. The tool will look up
the username, password and, optionally, the port in the conf file. The
example below shows how it works. If the port isn't specified it
defaults to the global (-P if specified or 22).

    EXAMPLE 04: Use the login conf file.

    $ # Create the conf file.
    $ cat >login.conf.txt <<EOF
    # login credentials
    {
      "global" : ["admin", "password", 22],
      "host" : {
          "host1" : ["admin", "password", 22],
          "host2" : ["admin", "password"],  # use the global default
      }
    }
    EOF
    $ # Now run the command.
    $ rcmd.py -l login.conf.txt -H host1,host2 -c pwd

The login conf file has two records: global and host. The global
record defines the login credentials to use for any host that is not
defined in the host section. The host section contains the login
credentials for specific hosts. The data in both cases is the
username, password and, optionally, the port.

Of course storing this data in plaintext is not very secure so the
tool provides a way to read encrypted data using the AES-CBC-256
encryption algorithm. The example below shows how to invoke the secure
version.

    EXAMPLE 05: Use an encrypted login conf file.

    $ # Encrypt the conf file.
    $ # It is okay to read from and write to the same file
    $ # because it buffers it internally.
    $ rcmd_crypt -e -p secret -i login.conf -o login.conf

    $ # Use the encrypted conf file.
    $ # If you specify an empty string, you will be prompted.
    $ rmcd.py -L login.conf '' -H host1,host2 -c pwd
    login.conf Password:

    $ # The same is true if use the -l option:
    $ rmcd.py -l login.conf -H host1,host2 -c pwd
    login.conf Password:

    $ # You can also specify the password on the command line:
    $ rmcd.py -L login.conf 'password' -H host1,host2 -c pwd

    $ # You can also specify the password on the command line
    $ # but you may want to cat it from a protected file to
    $ # avoid having it show up in the history log:
    $ rmcd.py -L login.conf "`cat password.file`" -H host1,host2 -c pwd

Once you have created a login conf file, you can use it to drive the
operation of the tool. It will operate on all of the hosts defined in
the conf file if no hosts are explicitly specified but you can control
that using filtering. The example below shows how this works with no
filtering.

    EXAMPLE 06: Use the login conf file to define the hosts.

    $ # Run on all hosts defined in the login conf file.
    $ rcmd.py -L login.conf -c pwd
    login.conf Password:

There are many cases when you only want to subset of the hosts. You
can limit the hosts specifying which hosts to include or by specifying
which ones to exclude. That is done by allowing regular expressions in
the host specification to define which hosts to include and by
recognizing a '!' prefix in a host specification to define which hosts
to exclude. This is what I meant by the term "filtering".

The example below shows how to specify which hosts to include. I have
beefed up the login.conf to have more host names.

    EXAMPLE 07: Choose a subset of hosts to run.

    $ # Create the login.conf.
    $ $ cat >login.conf <<EOF
    # login credentials
    {
      "global" : ["admin", "password", 22],  # 22 is the default
      "host" : {
          "host1" : ["admin", "password", 22],  # 22 is the default
          "host2" : ["admin", "password"],
          "filer1" : ["admin", "password"],
          "filer2" : ["admin", "password"],
          "ldap1" : ["admin", "password"],
          "ldap2" : ["admin", "password"],
      }
    }
    EOF

    $ # Encrypt it.
    $ # Note that you can encrypt/decrypt to the same file.
    $ rcmd_crypt -e -p secret -i login.conf -o login.conf

    $ # Operate on all hosts.
    $ rcmd.py -l login.conf -c pwd
    login.conf passsword:

    $ # Only operate on host1.
    $ rcmd.py -l login.conf -H host1 -c pwd
    login.conf passsword:

    $ # Only operate on host1 and host2.
    $ rcmd.py -l login.conf -H host1,host2 -c pwd
    login.conf passsword:

    $ # Only operate on the hosts whose names begin with 'h' and 'f'.
    $ rcmd.py -l login.conf -H '^[hf].*$' -c pwd
    login.conf passsword:

As you can see, this is very convenient but there are cases where you
want to exclude a small number of hosts which can be cumbersome using
the model above. To do that you can use the unary negation operator (!).
The example below shows how to do that.

    EXAMPLE 08: Exclude a subset of hosts.

    $ # All hosts except ldap2.
    $ rcmd.py -l login.conf.txt -H '!ldap2' -c pwd

    $ # All hosts except ldap1 and ldap2.
    $ rcmd.py -l login.conf.txt -H '!ldap1,ldap2' -c pwd

    $ # All hosts except ldap hosts.
    $ rcmd.py -l login.conf.txt -H '!^ldap[0-9]+$' -c pwd

That covers host specifications. We will now look at the command
options in more detail.

Multiple commands can be specified including sudo commands that
require input on each host. Furthermore, the default operation of the
tools only displays the output from the commands so you cannot even
tell which host the command is running on. This section describes how
to deal with all of these situations.

The example below shows how to run multiple commands.

    EXAMPLE 09: Multiple commands.

    $ # Add a header with the host name
    $ rcmd.py -l login.conf.txt \\
        -c 'echo && echo "=====" && echo "HOST: `hostname`" \\
        -c date \\
        -c uptime

The example below shows how to run a sudo command.

    EXAMPLE 10: sudo command

    $ # Use a sudo command.
    $ rcmd.py -l login.conf.txt \\
        -c 'echo && echo "=====" && echo "HOST: `hostname`" \\
        -c date \\
        -C 'sudo uname -a' 'password'

When you have multiple commands to run it is often easier to batch
them up by putting them in a single file and running them all at
once. This is done using the batch (-b) option as shown below.

    EXAMPLE 11: batch processing

    $ # Use a batch file.
    $ cat >batch.sh <<EOF
    #!/bin/bash
    echo '# ================================================================'
    echo "# BATCH TEST: `hostname`"
    echo '# ================================================================'
    echo "DIR: `pwd`"
    echo "DATE: `date`"
    echo "OS: `uname -a`"
    echo "RELEASE"
    lsb_release -a | sed -e 's/^/    /g'
    $ rcmd.py -l login.conf.txt -b batch.sh
EOF
'''
import argparse
import datetime
import logging
import paramiko
import os
import re
import socket
import sys
import time
import yaml

from rcmd_crypt import decrypt
from getpass import getpass
from pprint import pprint

# Globals
logging.basicConfig()
LOGGER = logging.getLogger('rcmd')
VERSION = '1.0'
COPYRIGHT = '''
    rcmd.py - executes commands on remote hosts
    Copyright (C) 2013  Joe Linoff

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''


# ================================================================
# class RunRemoteCmd
# ================================================================
class RunRemoteCmd:
    '''
    Create an SSH connection to a server and execute commands.
    Here is a typical usage:

        ssh = RunRemoteCmd()
        ssh.connect('host', 'user', 'password', port=22)
        if ssh.connected() is False:
            sys.exit('Connection failed')

        # Run a command that does not require input.
        status, output = ssh.run('uname -a')
        print 'status = %d' % (status)
        print 'output (%d):' % (len(output))
        print '%s' % (output)

        # Run a command that does requires input.
        status, output = ssh.run('sudo uname -a', 'sudo-password')
        print 'status = %d' % (status)
        print 'output (%d):' % (len(output))
        print '%s' % (output)
    '''
    def __init__(self, compress=True, verbose=False):
        '''
        Setup the initial verbosity level and the logger.

        @param compress  Enable/disable compression.
        @param verbose   Enable/disable verbose messages.
        '''
        self.ssh = None
        self.transport = None
        self.compress = compress
        self.bufsize = 65536

        # Setup the logger
        self.logger = logging.getLogger('RunRemoteCmd')
        self.set_verbosity(verbose)

        self.hostname = None
        self.username = None
        self.port = 22

        fmt = '%(asctime)s RunRemoteCmd:%(funcName)s:%(lineno)d %(message)s'
        formatter = logging.Formatter(fmt)
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.info = self.logger.info
        self.info('created object')

    def __del__(self):
        if self.transport is not None:
            self.transport.close()
            self.transport = None

    def connect(self, hostname, username, password, port=22):
        '''
        Connect to the host.

        @param hostname  The hostname.
        @param username  The username.
        @param password  The password.
        @param port      The port (default=22).

        @returns True if the connection succeeded or false otherwise.
        '''
        self.info('connecting %s@%s:%d' % (username, hostname, port))
        self.hostname = hostname
        self.username = username
        self.port = port
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            self.ssh.connect(hostname=hostname,
                             port=port,
                             username=username,
                             password=password)
            self.transport = self.ssh.get_transport()
            self.transport.use_compression(self.compress)
            self.info('succeeded: %s@%s:%d' % (username,
                                               hostname,
                                               port))
        except socket.error as exp:
            self.transport = None
            self.info('failed: %s@%s:%d: %s' % (username,
                                                hostname,
                                                port,
                                                str(exp)))
        except paramiko.BadAuthenticationType as exp:
            self.transport = None
            self.info('failed: %s@%s:%d: %s' % (username,
                                                hostname,
                                                port,
                                                str(exp)))
        except paramiko.AuthenticationException as exp:
            self.transport = None
            self.info('failed: %s@%s:%d: %s' % (username,
                                                hostname,
                                                port,
                                                str(exp)))

        return self.transport is not None

    def run(self, cmd, input_data=None, timeout=10):
        '''
        Run a command with optional input data.

        Here is an example that shows how to run commands with no input:

            ssh = RunRemoteCmd()
            ssh.connect('host', 'user', 'password')
            status, output = ssh.run('uname -a')
            status, output = ssh.run('uptime')

        Here is an example that shows how to run commands that require input:

            ssh = RunRemoteCmd()
            ssh.connect('host', 'user', 'password')
            status, output = ssh.run('sudo uname -a', '<sudo-password>')

        @param cmd         The command to run.
        @param input_data  The input data (default is None).
        @param timeout     The timeout in seconds (default is 10 seconds).
        @returns The status and the output (stdout and stderr combined).
        '''
        self.info('running command: (%d) %s' % (timeout, cmd))

        if self.transport is None:
            self.info('no connection to %s@%s:%s' % (str(self.username),
                                                     str(self.hostname),
                                                     str(self.port)))
            return -1, 'ERROR: connection not established\n'

        # Fix the input data.
        input_data = self._run_fix_input_data(input_data)

        # Initialize the session.
        self.info('initializing the session')
        session = self.transport.open_session()
        session.set_combine_stderr(True)
        session.get_pty()
        session.exec_command(cmd)
        output = self._run_poll(session, timeout, input_data)
        status = session.recv_exit_status()
        self.info('output size %d' % (len(output)))
        self.info('status %d' % (status))
        return status, output

    def connected(self):
        '''
        Am I connected to a host?

        @returns True if connected or false otherwise.
        '''
        return self.transport is not None

    def set_verbosity(self, verbose):
        '''
        Turn verbose messages on or off.

        @param verbose  Enable/disable verbose messages.
        '''
        if verbose > 0:
            self.logger.setLevel(logging.INFO)
        else:
            self.logger.setLevel(logging.ERROR)

    def _run_fix_input_data(self, input_data):
        '''
        Fix the input data supplied by the user for a command.

        @param input_data  The input data (default is None).
        @returns the fixed input data.
        '''
        if input_data is not None:
            if len(input_data) > 0:
                if '\\n' in input_data:
                    self.info('fixing user input')
                    # Convert \n in the input into new lines.
                    lines = input_data.split('\\n')
                    input_data = '\n'.join(lines)
            return input_data.split('\n')
        return []

    def _run_send_input(self, session, stdin, input_data):
        '''
        Send the input data.

        @param session     The session.
        @param stdin       The stdin stream for the session.
        @param input_data  The input data (default is None).
        '''
        if input_data is not None:
            self.info('session.exit_status_ready() %s' %
                      str(session.exit_status_ready()))
            self.info('stdin.channel.closed %s' % str(stdin.channel.closed))
            if stdin.channel.closed is False:
                self.info('sending input data')
                stdin.write(input_data)

    def _run_poll(self, session, timeout, input_data):
        '''
        Poll until the command completes.

        @param session     The session.
        @param timeout     The timeout in seconds.
        @param input_data  The input data.
        @returns the output
        '''
        interval = 0.1
        maxseconds = timeout
        maxcount = maxseconds / interval

        # Poll until completion or timeout
        # Note that we cannot directly use the stdout file descriptor
        # because it stalls at 64K bytes (65536).
        self.info('polling (%d, %d)' % (maxseconds, maxcount))
        input_idx = 0
        timeout_flag = False
        start = datetime.datetime.now()
        output = ''
        session.setblocking(0)
        while True:
            if session.recv_ready():
                data = session.recv(self.bufsize)
                output += data
                self.info('read %d bytes, total %d' % (len(data), len(output)))

                if session.send_ready():
                    # We received a potential prompt.
                    # In the future this could be made to work more
                    # like pexpect with pattern matching.
                    if input_idx < len(input_data):
                        data = input_data[input_idx] + '\n'
                        input_idx += 1
                        self.info('sending input data %d' % (len(data)))
                        session.send(data)

            self.info('session.exit_status_ready() = %s' %
                      (str(session.exit_status_ready())))
            if session.exit_status_ready():
                break

            # Timeout check
            ett = datetime.datetime.now() - start
            et_secs = (ett.days * 3600 * 24)
            et_secs += ett.seconds
            et_secs += (float(ett.microseconds) / 1000000.)
            self.info('timeout check %d %d' % (et_secs, maxseconds))
            if et_secs > maxseconds:
                self.info('polling finished - timeout')
                timeout_flag = True
                break
            time.sleep(0.200)

        self.info('polling loop ended')
        if session.recv_ready():
            data = session.recv(self.bufsize)
            output += data
            self.info('read %d bytes, total %d' % (len(data), len(output)))

        self.info('polling finished - %d output bytes' % (len(output)))
        if timeout_flag:
            self.info('appending timeout message')
            output += '\nERROR: timeout after %d seconds\n' % (timeout)
            session.close()

        return output


# ================================================================
# _login_global_update
# ================================================================
def _login_global_update(opts, logins, username, password, port, filename):
    '''
    Update the login global record.

    @param opts     The command line options.
    @param logins   The login credentials.
    @param username The username.
    @param password The password.
    @param port     The port number.
    @param filename The filename.
    @returns the updated login credentials.
    '''
    # Blindly overwrite the global data from the YAML file
    # but warn the user.
    _username = logins['global'][0]
    _password = logins['global'][1]
    _port = logins['global'][2]
    if _username is not None:
        if _username != username or _password != password or _port != port:
            msg = 'overwriting global login credentials: "%s" --> "%s" from %s'
            if opts.nowarn is False:
                LOGGER.warn(msg % (str(_username),
                                   str(username),
                                   filename))
            # Only overwrite if it changed.
            logins['global'] = [username, password, port, filename]
    else:
        logins['global'] = [username, password, port, filename]
    return logins


# ================================================================
# _login_host_update
# ================================================================
def _login_host_update(opts, logins, hostname, username, password,
                       port, filename):
    '''
    Update the login host record.

    @param opts     The command line options.
    @param logins   The login credentials.
    @param hostname The hostname.
    @param username The username.
    @param password The password.
    @param port     The port number.
    @param filename The filename.
    @returns the updated login credentials.
    '''

    # Allow the user to use the global defaults
    # by specifying empty strings.
    # Ex. -u user -p password --port 1234 -D newhost '' '' 0
    if username == '' and opts.user is not None:
        username = opts.user
    if password == '' and opts.password is not None:
        password = opts.password
    if port == 0 and opts.port is not None:
        port = int(opts.port)

    if hostname in logins['host']:
        _username = logins['host'][hostname][0]
        _password = logins['host'][hostname][1]
        if len(logins['host'][hostname]) > 2:
            _port = int(logins['host'][hostname][2])
        elif opts.port:
            _port = int(opts.port)
        else:
            _port = 22
        if _username != username or _password != password or _port != port:
            msg = 'overwriting login credentials for host: '
            msg += '%s, "%s" --> "%s" from %s'
            if opts.nowarn is False:
                LOGGER.warn(msg % (hostname, _username,
                                   username, filename))

            # Only update if it changed.
            logins['host'][hostname] = [username, password, port, filename]
    else:
        logins['host'][hostname] = [username, password, port, filename]
    return logins


# ================================================================
# _load_login_data
# ================================================================
def _load_login_data(opts, logins, filename, input_data):
    '''
    Load the login file data.

    This function can be used for plaintext or encrypted data.

    @param opts        The command line options.
    @param logins      The login credentials.
    @param filename    The input file name for error reporting.
    @param input_data  The input data. For encrypted files,
                       this is the plaintext.
    @returns the updated login credentials.
    '''
    if input_data is None:
        LOGGER.error('no data from file: %s' % (filename))
        sys.exit(1)

    # Note the serious lack of structure syntax checking.
    yaml_data = yaml.load(input_data)

    if not isinstance(yaml_data, dict):
        LOGGER.error('no valid data found in file: %s' % (filename))
        sys.exit(1)

    # Make sure that the high level structure is okay.
    found = False
    for key in yaml_data:
        if key not in ['global', 'host']:
            LOGGER.error('unrecognized key (%s): %s' % (key, filename))
            sys.exit(1)
        else:
            found = True
    if not found:
        LOGGER.warning('no login data found in %s' % (filename))

    # Process the global record:
    # Two options:
    #   { "global": ["user", "password", "port"] ... }
    #   { "global": ["user", "password"] ... }
    if 'global' in yaml_data:
        _global = yaml_data['global']
        if not isinstance(_global, list):
            LOGGER.error('global entry is not a list: %s' % (filename))
            sys.exit(1)
        if len(_global) < 2 or len(_global) > 3:
            LOGGER.error('bad global list length (%d): %s' %
                         (len(_global), filename))
            sys.exit(1)
        username = _global[0]
        password = _global[1]
        if len(_global) > 2:
            port = int(_global[2])
        else:
            port = int(opts.port)
        logins = _login_global_update(opts, logins,
                                      username, password, port,
                                      filename)

    # Process the host record:
    # { "host": {"host1": ["user", "password", port],
    #            "host2": ["user", "password"],  # port=22 is the default
    #           }}
    if 'host' in yaml_data:
        _host = yaml_data['host']
        if not isinstance(_host, dict):
            LOGGER.error('host entry is not a dict: %s' % (filename))
            sys.exit(1)

        # Blindly add in all of the hosts.
        # Overwrite hosts that are already there but
        # generate a warning during the update operation.
        for host in _host:
            hostrec = _host[host]
            if not isinstance(hostrec, list):
                LOGGER.error('host entry %s is not a list: %s' %
                             (host, filename))
                sys.exit(1)
            if len(hostrec) < 2 or len(hostrec) > 3:
                LOGGER.error('bad host %s list length (%d): %s' %
                             (host, len(_global), filename))
                sys.exit(1)

            username = hostrec[0]
            password = hostrec[1]
            if len(hostrec) > 2:
                port = int(hostrec[2])
            else:
                port = int(opts.port)
            logins = _login_host_update(opts, logins,
                                        host, username, password, port,
                                        filename)
    return logins


# ================================================================
# _login_file
# ================================================================
def _login_file(opts, logins):
    '''
    Load the login credentials from a plaintext YAML file.

    Do not overwrite the global credentials if they are set.

    @param opts   The command line options.
    @param logins The login credentials.
    @returns the updated login credentials.
    '''
    if opts.login is not None:
        for item in opts.login:
            ifile = item[0]
            ifp = open(ifile, 'r')
            data = ifp.read()
            ifp.close()

            # Heuristic guess to determine whether this is an
            # encrypted file.
            if data is not None:
                if data.find('{') < 0:
                    # This isn't a YAML file.
                    # assume that is it encrypted.
                    password = getpass('%s password: ' % (ifile))
                    try:
                        data = decrypt(password, data)
                    except Exception as exp:  # pylint: disable=W0703
                        LOGGER.error('decryption failed for file: %s (%s)' %
                                     (ifile, str(exp)))
                        sys.exit(1)

            logins = _load_login_data(opts, logins, ifile, data)

    return logins


# ================================================================
# _login_file_encrypted
# ================================================================
def _login_file_encrypted(opts, logins):
    '''
    Load the login credentials from an encrypted YAML file.

    Encryption is done using openssl enc -e -a -aes-256-cbc -salt.

    Do not overwrite the global credentials if they are set.

    @param opts   The command line options.
    @param logins The login credentials.
    @returns the updated login credentials.
    '''
    if opts.login_secure is not None:
        for item in opts.login_secure:
            ifile = item[0]
            password = item[1]
            ifp = open(ifile, 'r')
            ciphertext = ifp.read()
            ifp.close()

            if password == '':
                password = getpass('%s password: ' % (ifile))
            try:
                plaintext = decrypt(password, ciphertext)
            except Exception as exp:  # pylint: disable=W0703
                LOGGER.error('decrypt failed for file: %s (%s)' %
                             (ifile, str(exp)))
                sys.exit(1)
            logins = _load_login_data(opts, logins, ifile, plaintext)

    return logins


# ================================================================
# _login_cli
# ================================================================
def _login_cli(opts, logins):
    '''
    Load the login credentials from the CLI: -u and -p.

    @param opts   The command line options.
    @param logins The login credentials.
    @returns the updated login credentials.
    '''
    if opts.user is not None:
        username = opts.user
        if opts.password:
            password = opts.password
        else:
            password = getpass(opts.user + ' password: ')
        if opts.port:
            port = int(opts.port)
        else:
            port = 22
        logins = _login_global_update(opts, logins,
                                      username, password, port,
                                      'CLI')
    elif opts.password is not None:
        LOGGER.error('password specified with no user name')
        sys.exit(1)
    return logins


# ================================================================
# _login_defined_hosts
# ================================================================
def _login_defined_hosts(opts, logins):
    '''
    Load the login credentials from the -D host specifications.
    These are the highest priority.

    @param opts   The command line options.
    @param logins The login credentials.
    @returns the updated login credentials.
    '''
    if opts.define is not None:
        for host in opts.define:
            logins = _login_host_update(opts, logins,
                                        host[0], host[1], host[2],
                                        int(host[3]),
                                        'DEFINE')

    return logins


# ================================================================
# _load_login_credentials
# ================================================================
def _login_credentials(opts):
    '''
    Load the login credentials.

    @param opts  The command line options.
    @returns the login credentials.
    '''
    logins = {'global': [None, None, None],
              'host': {}}
    logins = _login_file(opts, logins)
    logins = _login_file_encrypted(opts, logins)
    logins = _login_cli(opts, logins)
    logins = _login_defined_hosts(opts, logins)

    if logins['global'][0] is None and len(logins['host']) == 0:
        LOGGER.error('no login credentials specified')
        sys.exit(1)

    return logins


# ================================================================
# _load_hosts
# ================================================================
def _load_hosts(opts, logins):
    '''
    Load the host information.

    This function loads all of the host information then filters
    out information provide in regular expressions or leading
    dashes.

    All of the host specifications are in the login data structure.
    It contains host names, regexs and negatives (-<host>).

    @param opts   The command line options.
    @param logins The login credentials.
    @returns the updated login credentials.
    '''

    # If no hosts were specified on the command line, then
    # use all hosts.
    hosts = {}  # use a dictionary for fast filtering lookup
    excluded_hosts = []
    regex_hosts = []
    regex_excluded_hosts = []
    if opts.host is not None:
        for hostnames in opts.host:
            # Handle the case where multiple hostnames are
            # specified in a comma separated list:
            #    -H host1,host2,host3
            # also
            #    -H '!host1,host2,host3'
            negate = False
            if hostnames[0] == '!':
                negate = True
                hostnames = hostnames[:1]
            hostnames_list = hostnames.split(',')
            for host in hostnames_list:
                # Check to see whether this host is an re.
                # If it is store it for later.
                match = re.search('[\*\[\^\$]', host)
                if match:
                    if negate:
                        # Leading dash means exclude.
                        regex_excluded_hosts.append(host)
                    else:
                        regex_hosts.append(host)
                else:
                    if host in logins['host']:
                        username = logins['host'][host][0]
                        password = logins['host'][host][1]
                        port = int(logins['host'][host][2])
                    else:
                        username = logins['global'][0]
                        password = logins['global'][1]
                        port = int(logins['global'][2])
                        if username is None:
                            msg = 'invalid login credentials for host: %s'
                            LOGGER.error(msg % (host))
                            sys.exit(1)
                    if negate:
                        excluded_hosts.append(host)
                    else:
                        hosts[host] = [username, password, port]

    for regex_host in regex_hosts:
        # Search for matches in the login hosts that are not already
        # in the list.
        for loghost in logins['host']:
            if loghost not in hosts:
                match = re.search(regex_host, loghost)
                if match:
                    username = logins['host'][loghost][0]
                    password = logins['host'][loghost][1]
                    port = int(logins['host'][loghost][2])
                    hosts[loghost] = [username, password, port]

    # At this point, if no hosts have been added, use all of the login
    # hosts.
    if len(hosts) == 0:
        for loghost in logins['host']:
            username = logins['host'][loghost][0]
            password = logins['host'][loghost][1]
            port = int(logins['host'][loghost][2])
            hosts[loghost] = [username, password, port]

    # Now remove the hosts that were specifically excluded.
    if len(hosts):
        for excluded_host in excluded_hosts:
            if excluded_host in hosts:
                del hosts[excluded_host]

    # Now remove the hosts that were specifically excluded using
    # regex's.
    if len(hosts):
        for regex_host in regex_excluded_hosts:
            hosts_to_delete = []
            for host in hosts:
                match = re.search(regex_host, host)
                if match:
                    # Queue the hosts to delete.
                    # We don't want to disrupt the loop.
                    hosts_to_delete.append(host)

            # Now delete them.
            for host in hosts_to_delete:
                del hosts[host]

    return hosts


# ================================================================
# _load_cmds
# ================================================================
def _load_cmds(opts):
    '''
    Load the commands while preserving order.

    There are 3 types of commands: simple, simple + input and batch
    (read from a file).

    This function converts them two, a tuple: the command and the
    input that is used by the _runcmd function.

    @param opts     The command line options.
    @returns The command list.
    '''
    cmds = []
    if opts.cmd is not None:
        for cmd in opts.cmd:
            if isinstance(cmd, str):
                cmdn = cmd
                indata = None
            elif isinstance(cmd, list):
                assert len(cmd) in [1, 2]
                if len(cmd) == 1:
                    cfp = open(cmd[0], 'r')
                    cmdn = cfp.read()
                    cfp.close()
                    indata = None
                elif len(cmd) == 2:
                    cmdn = cmd[0]
                    indata = cmd[1]
            else:
                assert isinstance(cmd, str) or isinstance(cmd, list)
            cmds.append([cmdn, indata])

    return cmds


# ================================================================
# runcmd
# ================================================================
def _runcmd(opts, cmd, rcmd, hostname, port, username):
    '''
    Run a single command.

    @param opts     The command line options.
    @param cmd      The command to run with input (2 element list).
    @param rcmd     The remote command execution object.
    @param hostname The hostname.
    @param port     The port.
    @param username The username.
    '''
    if opts.verbose:
        print
        print '# %s' % ('=' * 64)
        print '# Host    : %s' % (hostname)
        print '# Port    : %d' % (port)
        print '# User    : %s' % (username)
        print '# Command : %s' % (cmd)
        print '# Timeout : %d' % (opts.timeout)
        start = datetime.datetime.now()

    output = 'ERROR: could not connect to the server: %s\n' % (hostname)
    status = -1
    if rcmd.connected():
        status, output = rcmd.run(cmd[0], cmd[1], timeout=float(opts.timeout))

    if opts.verbose:
        elapsed = datetime.datetime.now() - start
        elapsed_secs = elapsed.days * 3600 * 24 + elapsed.seconds
        elapsed_secs += float(elapsed.microseconds) / 1000000.
        print '# Status  : %d' % (status)
        print '# Size    : %d' % (len(output))
        print '# Time    : %.2f secons' % (elapsed_secs)
        print '# %s' % ('=' * 64)

    sys.stdout.write('%s' % (output))
    sys.stdout.flush()

    return status, output


# ================================================================
# _runcmds
# ================================================================
def _runcmds(opts, hosts):
    '''
    Run commands for all of the hosts.

    @param opts   The command line options.
    @param hosts  The hosts to operate on.
    '''
    cmds = _load_cmds(opts)
    if len(cmds):
        for host in sorted(hosts, key=str.lower):
            username = hosts[host][0]
            password = hosts[host][1]
            port = hosts[host][2]

            rcmd = RunRemoteCmd()
            if opts.verbose > 1:
                rcmd.set_verbosity(1)

            rcmd.connect(host, username, password, port)

            for cmd in cmds:
                _runcmd(opts, cmd, rcmd, host, port, username)


# ================================================================
# _get_opts
# ================================================================
def _get_opts():
    '''
    Define and parse the command line options.

    Use the pydoc command to get the module documentation and
    incorporate it into the help.

    @returns The options.
    '''
    # Tricky way to get the package description into the help.
    import pydoc
    import rcmd  # pylint: disable=W0406
    description = pydoc.getdoc(rcmd)
    epilog='\nCOPYRIGHT NOTICE\n\n%s' % (COPYRIGHT)
    rawd = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(formatter_class=rawd,
                                     description=description,
                                     epilog=epilog)
    parser.add_argument('-b', '--batch',
                        action='append',
                        nargs=1,  # append as a list
                        type=str,
                        dest='cmd',
                        help='batch commands')
    parser.add_argument('-c', '--cmd',
                        action='append',
                        dest='cmd',
                        help='a command to execute')
    parser.add_argument('-C', '--cmdi',
                        action='append',
                        dest='cmd',
                        nargs=2,
                        metavar=('CMD', 'INPUT'),
                        help='a command to execute with input')
    parser.add_argument('-D', '--define',
                        action='append',
                        nargs=4,
                        metavar=('HOST', 'USER', 'PASSWORD', 'PORT'),
                        help='define host data on the command line')
    parser.add_argument('-H', '--host',
                        action='append',
                        help='host or comma separated list of hosts')
    parser.add_argument('-l', '--login',
                        action='append',
                        nargs=1,
                        metavar=('FILE'),
                        help='login credentials file')
    parser.add_argument('-L', '--login-secure',
                        action='append',
                        nargs=2,
                        metavar=('ENCRYPTED_FILE', 'PASSWORD'),
                        help='secure login file with password')
    parser.add_argument('--nowarn',
                        action='store_true',
                        help='disable warnings')
    parser.add_argument('-p', '--password',
                        action='store',
                        help='the default user password')
    parser.add_argument('-P', '--port',
                        type=int,
                        default=22,
                        action='store',
                        help='the SSH port, default is 22')
    parser.add_argument('-t', '--timeout',
                        action='store',
                        type=float,
                        default=300,
                        metavar=('SECONDS'),
                        help='the command timeout, default is 300')
    parser.add_argument('-u', '--user',
                        action='store',
                        help='the default user name, overrides the login DB')
    parser.add_argument('-v', '--verbose',
                        action='count',
                        help='the level of verbosity')
    parser.add_argument('-V', '--version',
                        action='version',
                        version='%(prog)s ' + VERSION)

    args = parser.parse_args()

    if args.verbose:
        print '''%s  Copyright(C) 2013  Joe Linoff
This program comes with ABSOLUTELY NO WARRANTY; for details type visit
http://www.gnu.org/licenses/.
This is free software, and you are welcome to redistribute it under
certain conditions; see the above web site for details.
''' % os.path.basename(__file__)

    if args.cmd is None:
        LOGGER.error('at least one command (-c or -C) must be specified')
        sys.exit(1)

    for cmd in args.cmd:
        if isinstance(cmd, list) and len(cmd) == 1:
            if not os.path.exists(cmd[0]):
                LOGGER.error('batch file does not exist: %s' % (cmd[0]))
                sys.exit(1)

    if args.verbose > 2:
        pprint(args, indent=4)
    return args


# ================================================================
# main
# ================================================================
def main():
    '''
    Run the program.
    '''
    opts = _get_opts()
    logins = _login_credentials(opts)
    hosts = _load_hosts(opts, logins)
    _runcmds(opts, hosts)

# ================================================================
# MAIN
# ================================================================
if __name__ == '__main__':
    main()
